**Table Of Contents**
___

[[_toc_]]
___

# Python on GPU with JAX
## JAX on GPU

JAX is made for computing on CPU, GPU or TPU without having to modify one's code. Here are a few tips to know when you use JAX on your GPU though:
- you have to install JAX with [support for GPU](https://jax.readthedocs.io/en/latest/installation.html), whether it be NVIDIA, AMD, ARM-based GPU or TPU.
- it is recommended to use optimization flags when using a GPU with JAX: [XLA optimization flags] https://jax.readthedocs.io/en/latest/gpu_performance_tips.html#xla-performance-flags, for example:
```python
import os
os.environ['XLA_FLAGS'] = (
    '--xla_gpu_enable_triton_softmax_fusion=true '
    '--xla_gpu_triton_gemm_any=True '
)
```
Apart from that, you should be able to use the same JAX code on a CPU or on a GPU !

## Tutorial : Gray-Scott on GPU with JAX
We will use the [gray_scott_jax.py](gray_scott_jax.py) for this tutorial. It contains the Gray-Scott code optimized with JAX which we produced for CPU on our Jupyter notebook used previously.

Complete the code to add performance flags at the top of the file and to use the Gray-Scott functions in the `main`.

Then, run the benchmark on the GPU for 34x1000 images.

## Exercise : Use [scan](https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.scan.html) to optimize

For this exercise we are going to use a simplified Gray-Scott code, computing 1000 images but with `dsteps` set to 1. We will optimize the outer loop on images. We set `dsteps` to 1 because the optimization we are goig to perform will interact badly with an inner loop on steps, so we remove the inner loop to simplify.

If we want to optimize with JAX our main function, we are going to have to get rid of the loops, like we did with the laplacian. The inner loop on dsteps will be totally removed as we consider in this exercise that `dsteps=1`. Then, for the outer loop we can use [jax.lax.scan](https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.scan.html) to scan a function computing `V` over all the images contained in `images_array`, while carrying along the state of `V` for the next image. This step is a bit tricky but thanks to it we can use the full power of JAX.

To do so, let's define a function called `compute_image`. Accoording to the doc, `compute_image` must be a Python function that accepts two arguments where the first is a value of the loop carry and the second is a slice of `images_array` along its leading axis. It returns a pair where the first element represents a new value for the loop carry and the second represents a slice of the output.
In our case, the only argument we need is the loop carry.

We are going to use the [gray_scott_jax_scan.py](gray_scott_jax_scan.py) file and fill the required functions.

If you want to assess that your final code is indeed faster with this version, you can take [gray_scott_jax.py](gray_scott_jax.py), set `dsteps` to 1 and rerun it to get the execution time. 