**Table Of Contents**
___

[[_toc_]]
___

# Python on GPU with cuNumeric
## Legate
**Legate** is an Nvidia project aimed at democratizing distributed and heterogeneous computing on CPU and GPUs. It makes parallelization implicit, so that the developer does not have to worry about it.
It is based on the **Legion** programming system and on the **Realm** runtime system (based itself on CUDA).

![img](img/Legate-software-stack-ecosystem.png)
(*From [Nvidia Developer](https://developer.nvidia.com/blog/accelerating-python-applications-with-cunumeric-and-legate/)*)

Two librairies based on Legate are available for Python programming : `cuNumeric` and `legate.pandas`.

## cuNumeric
[cuNumeric](https://docs.nvidia.com/cunumeric/24.06/index.html) is a drop-in Numpy replacement. It covers more than 60% of the Numpy API. The latest version of the code was just released, on the 4th of July 2024. 

To install it, you need Python >= 3.10, conda >= 24.1 and CUDA >= 11.8.

To use it, just replace Numpy with it :
`import cunumeric as np`

To execute your program, run it with the legate driver : 
`legate --cpus 2 --gpus 2 program.py`

In principle, you should not have to make other adjustments. Keep in mind that you will have better performances if your code is vectorized, uses arrays and avoids loops (see https://docs.nvidia.com/cunumeric/24.06/user/practices.html).

cuNumeric performs two types of **implicit parallelization**:
- on **data** : it splits the data into subsets and Legion sends subsets to the processors
- on **tasks** : Legion produces a **dependency graph** of the tasks and handles their asynchronous execution.

![async](img/asynchronous-execution-cuNumeric.jpg)
(*From [Nvidia Developer](https://developer.nvidia.com/blog/accelerating-python-applications-with-cunumeric-and-legate/)*)

## Feedback and testing

- v.23 had a lot of bugs, especially with array slicing. The error messages were not helping. I had to rewrite Gray-Scott to have a version that would run with cuNumeric, and it was very slow to execute (slower than on a CPU).
- v.24 has less bugs but it was released only a few days ago and the installation process is not easy. 
It is a rather young project (beta release was made 1 year ago), still in development, but it looks promising.

We ran the [gray_scott_cunumeric.py](gray_scott_cunumeric.py) program on a GPU. On an A100, it takes 35s for only 34x100 steps, so it is very slow, even if we rewrote the Gray-Scott program to follow most of the [best practices for cuNumeric](https://nv-legate.github.io/cunumeric/23.11/user/practices.html). 